/*
   Class for controlling all of the devices (pots/switches) that
   control the HMSynth keyboard.

   Concept, Design and Implementation by: Craig A. Lindley
   Last Update: 04/26/2022
*/

#ifndef CONTROLS_H
#define CONTROLS_H

#include "MultiButton.h"
#include "Pot.h"
#include "RotaryEncoder.h"

// Threshold value used to determine if Pots have changed values
#define DELTA_THRESHOLD 5

// GPIO for pushbuttons and rotary encoder
#define MF_Pin        24 // Multi function pushbutton pin
#define Up_Pin        25
#define Down_Pin      26
#define RE_Plus_Pin   27
#define RE_Minus_Pin  28
#define RE_Switch_Pin 29

// Analog connections for pots and joystick
#define Mod_Pin      A14 // white
#define Pitch_Pin    A15 //green
#define Pot1_Pin     A16
#define Pot2_Pin     A17
#define Pot3_Pin     A18
#define Pot4_Pin     A19

// Instantiate Rotary Encoder
RotaryEncoder encoder = RotaryEncoder(RE_Plus_Pin, RE_Minus_Pin);

// Instantiate Rotary Encoder pushbutton for possible multi click operation
PinButton encoderBTN = PinButton(RE_Switch_Pin);

// Instantiate pushbuttons for possible multi click operation
PinButton mfBTN = PinButton(MF_Pin);
PinButton upBTN = PinButton(Up_Pin);
PinButton downBTN = PinButton(Down_Pin);

// Instantiate the 6 Pots used by this MIDI controller
Pot modWheel   = Pot(Mod_Pin,   15); // These values different because of noisy joystick readings
Pot pitchWheel = Pot(Pitch_Pin, 15);
Pot pot1 = Pot(Pot1_Pin,  DELTA_THRESHOLD);
Pot pot2 = Pot(Pot2_Pin,  DELTA_THRESHOLD);
Pot pot3 = Pot(Pot3_Pin,  DELTA_THRESHOLD);
Pot pot4 = Pot(Pot4_Pin,  DELTA_THRESHOLD);

class Controls {

  public:

    // Class constructor
    Controls() {}

    // Update the buttons' states
    void controlsUpdate(void) {
      mfBTN.update();
      upBTN.update();
      downBTN.update();
      encoderBTN.update();
    }

    // Multi function button functions
    boolean multiFunctionButtonSingleClicked(void) {
      return mfBTN.isSingleClick();
    }

    boolean multiFunctionButtonDoubleClicked(void) {
      return mfBTN.isDoubleClick();
    }

    boolean multiFunctionButtonLongClicked(void) {
      return mfBTN.isLongClick();
    }

    // Up button functions
    boolean upButtonSingleClicked(void) {
      return upBTN.isSingleClick();
    }

    boolean upButtonDoubleClicked(void) {
      return upBTN.isDoubleClick();
    }

    boolean upButtonLongClicked(void) {
      return upBTN.isLongClick();
    }

    // Down button functions
    boolean downButtonSingleClicked(void) {
      return downBTN.isSingleClick();
    }

    boolean downButtonDoubleClicked(void) {
      return downBTN.isDoubleClick();
    }

    boolean downButtonLongClicked(void) {
      return downBTN.isLongClick();
    }

    // Encoder button functions
    boolean encoderButtonSingleClicked(void) {
      return encoderBTN.isSingleClick();
    }

    boolean encoderButtonDoubleClicked(void) {
      return encoderBTN.isDoubleClick();
    }

    boolean encoderButtonLongClicked(void) {
      return encoderBTN.isLongClick();
    }

    // Read the rotary encoder and update value variable accordingly
    boolean readEncoder(int32_t *value, int inc,
                        int32_t minValue, int32_t maxValue) {

      // Assume encoder value changed and value will be returned
      int returnValue = true;

      int result = encoder.pollEncoder();
      if (result == DIR_CW) {
        *value += inc;
        if (*value > maxValue) {
          *value = maxValue;
        }
      } else if (result == DIR_CCW) {
        *value -= inc;
        if (*value < minValue) {
          *value = minValue;
        }
      } else {
        returnValue = false;
      }
      return returnValue;
    }

    // Potentiometer functions
    boolean modWheelActive(void) {
      return modWheel.isActive();
    }

    uint16_t modWheelValue(void) {
      return modWheel.getValue(true);
    }

    boolean pitchWheelActive(void) {
      return pitchWheel.isActive();
    }

    uint16_t pitchWheelValue(void) {
      return pitchWheel.getValue(true);
    }

    boolean pot1Active(void) {
      return pot1.isActive();
    }

    uint16_t pot1Value(void) {
      return pot1.getValue(false);
    }

    boolean pot2Active(void) {
      return pot2.isActive();
    }

    uint16_t pot2Value(void) {
      return pot2.getValue(false);
    }

    boolean pot3Active(void) {
      return pot3.isActive();
    }

    uint16_t pot3Value(void) {
      return pot3.getValue(false);
    }

    boolean pot4Active(void) {
      return pot4.isActive();
    }

    uint16_t pot4Value(void) {
      return pot4.getValue(false);
    }
};

#endif
