/*
   Class for mangeing the TFT on the HMSynth keyboard

   LCD Dimensions: 160x128
   Text Size = 1, 16 lines, 25 chars, ch height 8
   Text Size = 2, 8 lines, 13 chars, ch height 16
   Text Size = 3, 5 lines, 9 chars, ch height 24
   Colors that show OK: white, green and yellow
   Other colors too dim

   Concept, Design and Implementation by: Craig A. Lindley
   Last Update: 04/20/2022
*/

#ifndef DISPLAYMANAGER_H
#define DISPLAYMANAGER_H

#include <Adafruit_GFX.h>
#include <Adafruit_ST7735.h>
#include <SPI.h>

// TFT LCD signal connections to Teensy 3.6
#define TFT_DC    9 // brown
#define TFT_CS   10 // yellow
#define TFT_MOSI 11 // white/orange
#define TFT_SCLK 13 // orange
#define TFT_RST  -1
//      TFT_VCC     // red
//      TFT_GND     // gray

// Instantiate the LCD driver
Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCLK, TFT_RST);

class DisplayManager {

  public:

    // Class constructor
    DisplayManager() {}

    void begin(void) {

      // Initializer for 1.8" TFT screen
      tft.initR(INITR_BLACKTAB);
      // Set rotation of display
      tft.setRotation(3);
      // Clear the screen
      tft.fillScreen(ST77XX_BLACK);
    }

    void clearScreen(void) {

      // Clear the screen
      tft.fillScreen(ST77XX_BLACK);
    }

    // Draw text on display.
    // line numbers start at 0; tSize is 1, 2, 3 or 4
    void drawText(const char *text, int x, int line, int tSize, uint16_t tColor) {

      tft.setTextColor(tColor);
      tft.setTextSize(tSize);
      tft.setTextWrap(false);

      tft.setCursor(x, (line * 8 * tSize) + 3);
      tft.print(text);
    }

    // Draw text centered on line
    void drawCenteredText(const char *text, int line, int tSize, uint16_t tColor) {

      int16_t x1 = 0;
      int16_t y1 = 0;
      int16_t x2 = 0;
      int16_t y2 = 0;
      uint16_t w, h;

      tft.setTextSize(tSize);
      tft.getTextBounds(text, x1, y1, &x2, &y2, &w, &h);

      drawText(text, (tft.width() - w) / 2, line, tSize, tColor);
    }

    // Draw right justified text on line
    void drawRightJustifiedText(const char *text, int line, int tSize, uint16_t tColor) {

      int16_t x1 = 0;
      int16_t y1 = 0;
      int16_t x2 = 0;
      int16_t y2 = 0;
      uint16_t w, h;

      tft.setTextSize(tSize);
      tft.getTextBounds(text, x1, y1, &x2, &y2, &w, &h);

      drawText(text, tft.width() - w, line, tSize, tColor);
    }

    // Format LCD for subsequent info display
    void formatLCD(void) {

      // Clear the screen
      clearScreen();

      // Draw the border
      tft.drawRect(0, 0, tft.width(), tft.height(), ST77XX_BLUE);

      // Draw static text
      drawCenteredText("HMSynth", 0, 2, ST77XX_WHITE);
      drawCenteredText("Keyboard", 1, 2, ST77XX_WHITE);
      drawCenteredText("Controller", 2, 2, ST77XX_WHITE);

      // Format version string
      sprintf(textBuffer, "Ver: %d.%d", MAJ_VERSION, MIN_VERSION);

      drawCenteredText("by", 4, 2, ST77XX_GREEN);
      drawCenteredText("Craig Lindley", 5, 2, ST77XX_GREEN);
      drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
    }

    // Erase the info area of the screen. Assumes tSize is 2
    void eraseInfoAreaFromLine(int line) {

      // Calculate Y value
      int y = (line * 8 * 2) + 3;

      // Clear area
      tft.fillRect(1, y, tft.width() - 3, tft.height() - y - 1, ST77XX_BLACK);
    }

  protected:

    char textBuffer[30];
};

#endif
