/*
  Keyboard Polling Class

  This code polls the Yamaha keyboard in a row/column fashion
  and generates KEYEVENTs detailing which key numbers 1..36 were
  active and whether they are pressed or released.

   Ribbon Cable Connections from Yamaha Keyboard

   Pin           Signal      Teensy 3.6 GPIO Pin
   =============================================
   1 (purple)     nc
   2              nc
   3              nc
   4              R6                8
   5              nc
   6              R5                7
   7              C5               23
   8              R4                6
   9              C4               22
   10             R3                5
   11             C3               21
   12             R2                4
   13             C2               20
   14             R1                3
   15             C1               19
   16             R0                2
   17             C0               18

      Row key connections
   Row              Key Number
   R0               1
   R1               2, 3, 4, 5, 6, 7
   R2               8, 9,10,11,12,13
   R3              14,15,16,17,18,19
   R4              20,21,22,23,24,25
   R5              26,27,28,29,30,31
   R6              32,33,34,35,36,X

  Concept, Design and Implementation: Craig A. Lindley
  Last Update: 04/20/2022
*/

#ifndef KEYBOARD_H
#define KEYBOARD_H

// Define the keyboard event function
typedef void (*KEYEVENT)(int keyState, int keyNumber);

// Keyboard Row GPIO definitions
#define Row0 2
#define Row1 3
#define Row2 4
#define Row3 5
#define Row4 6
#define Row5 7
#define Row6 8

// Keyboard Column GPIO definitions
#define Col0 18
#define Col1 19
#define Col2 20
#define Col3 21
#define Col4 22
#define Col5 23

// Mappings from row and col numbers to GPIO pin numbers
int rowToPin[] = {Row0, Row1, Row2, Row3, Row4, Row5, Row6};
int colToPin[] = {Col0, Col1, Col2, Col3, Col4, Col5};

// Key matrix
// Keys numbered 1..36
int matrix[7][6] = {
  0,  0,  0,  0,  0,  1,
  2,  3,  4,  5,  6,  7,
  8,  9, 10, 11, 12, 13,
  14, 15, 16, 17, 18, 19,
  20, 21, 22, 23, 24, 25,
  26, 27, 28, 29, 30, 31,
  32, 33, 34, 35, 36, -1
};

// 64 bit bit field macros for key history
#define bitTest64(value, bit) (((value) >> (bit)) & 0x01)
#define bitSet64(value, bit) ((value) |= (1ULL << (bit)))
#define bitClear64(value, bit) ((value) &= ~(1ULL << (bit)))

// Keyboard class
class Keyboard {

  public:

    // Class constructor
    Keyboard() {

      // Setup GPIO pins for keyboard column inputs
      pinMode(Col0, INPUT_PULLUP);
      pinMode(Col1, INPUT_PULLUP);
      pinMode(Col2, INPUT_PULLUP);
      pinMode(Col3, INPUT_PULLUP);
      pinMode(Col4, INPUT_PULLUP);
      pinMode(Col5, INPUT_PULLUP);

      // Clear keyHistory bit field
      keyHistory = 0;
    }

    // Retrieve the key state
    // Returns true if key is on; false otherwise
    boolean getKeyState(int keyNumber) {
      return (bitTest64(keyHistory, keyNumber) != 0);
    }

    // Clear all key history
    void clearKeyHistory(void) {
      keyHistory = 0;
    }

    // Poll keyboard and generate KEYEVENTs for each active key
    void pollKeyboard(KEYEVENT ke) {

      // Check each row
      for (int row = 0; row < 7; row++) {

        // Set that row to OUTPUT and LOW
        pinMode(rowToPin[row], OUTPUT);
        digitalWrite(rowToPin[row], LOW);
        delayMicroseconds(50);

        // Check each column to see if a key has driven that column LOW
        for (byte col = 0; col < 6; col++) {

          // Get key number associated with row/col
          int keyNumber = matrix[row][col];

          if (keyNumber == -1) {
            continue;
          }

          // Is this key active ? keyState true says yes
          bool keyState = digitalRead(colToPin[col]) == LOW;

          // Has key state changed ?
          if (keyState != (bitTest64(keyHistory, keyNumber) != 0)) {
            // State has changed, log new state in bit field
            if (keyState) {
              bitSet64(keyHistory, keyNumber);
            } else  {
              bitClear64(keyHistory, keyNumber);
            }
            // Fire a key event for the active key
            // Serial.printf("Key: %d %s\n", keyNumber, keyState ? "down" : "up");
            ke(keyState, keyNumber);
          }
        }
        // Drive row back high and to high-impedance input state
        digitalWrite(rowToPin[row], HIGH);
        pinMode(rowToPin[row], INPUT);
        delayMicroseconds(50);
      }
    }

  protected:

    uint64_t keyHistory;
};

#endif
