/*
   HMSynth Keyboard Controller

   Concept, Design and Implementation by: Craig A. Lindley
   Last Update: 04/20/2022
*/

// Software version numbers
#define MAJ_VERSION 0
#define MIN_VERSION 6

// Define the following to enable USB MIDI output
// Must set Tools/USB Type to MIDI as well
//#define ENABLE_USB_MIDI

// Define defaults
#define DEFAULT_MIDI_CHANNEL         1
#define DEFAULT_MIDI_VELOCITY      100
#define DEFAULT_MIDI_NOTE_OFFSET    48 // Place keyboard at C3
#define DEFAULT_KEYBOARD_SHIFT       0 // 0 shift keyboard at C3
#define DEFAULT_MIDI_PATCH           1
#define DEFAULT_CCPOT_1              5 // Portamento time
#define DEFAULT_CCPOT_2              7 // Volume
#define DEFAULT_CCPOT_3             10 // Pan
#define DEFAULT_CCPOT_4             91 // Reverb send
#define PORTAMENTO_CC               65 // Portamento on/off

// Split definition
typedef struct {
  int channel;
  int velocity;
  int keyboardShift;
  int patch;
} SPLIT;

// Three splits of the keyboard are available
// Each with its own parameters
SPLIT splits[3];

#include "Controls.h"
#include "Keyboard.h"
#include "DisplayManager.h"
#include "MIDIIntfc.h"
#include "ChannelSelector.h"
#include "VelocitySelector.h"
#include "KeyShiftSelector.h"
#include "PatchSelector.h"
#include "StateMachine.h"

/*********************************************************/
/***               Instantiate all drivers             ***/
/*********************************************************/

// Instantiate the Controls class
Controls controls = Controls();

// Instantiate the Keyboard class
Keyboard kb = Keyboard();

// Instantiate the Display Manager class
DisplayManager dm = DisplayManager();

// Instantiate the MIDI Interface class
MIDIIntfc mi = MIDIIntfc();

// Instantiate MIDI channel selector
ChannelSelector cs = ChannelSelector();

// Instantiate MIDI velocity selector
VelocitySelector vs = VelocitySelector();

// Instantiate MIDI key shift selector
KeyShiftSelector kss = KeyShiftSelector();

// Instantiate MIDI patch selector
PatchSelector ps = PatchSelector();

// Instantiate the finite State Machine class
StateMachine fsm = StateMachine(controls, dm, kb, mi, cs, vs, kss, ps);

/*********************************************************/
/***                    Misc Functions                 ***/
/*********************************************************/

// Process key event
void processKeyboardEvent(int keyState, int keyNumber) {

  int channel;
  int velocity;
  int keyboardShift;

  // Determine split region from keyNumber
  if ((keyNumber >= 1) and (keyNumber <= 12)) {
    activeSplit = 0;
  } else if ((keyNumber >= 13) and (keyNumber <= 24)) {
    activeSplit = 1;
  } else {
    activeSplit = 2;
  }

  // Keyboard splits enabled
  switch (activeSplit) {
    case 0:
      channel = splits[0].channel;
      velocity = splits[0].velocity;
      keyboardShift = splits[0].keyboardShift;
      break;

    case 1:
      channel = splits[1].channel;
      velocity = splits[1].velocity;
      keyboardShift = splits[1].keyboardShift;
      break;

    case 2:
      channel = splits[2].channel;
      velocity = splits[2].velocity;
      keyboardShift = splits[2].keyboardShift;
      break;
  }
  // Do MIDI outout
  mi.doMidiOutput(channel, velocity, keyboardShift, keyState, keyNumber);
}

/*********************************************************/
/***                    Program Setup                  ***/
/*********************************************************/
void setup(void) {

  Serial.begin(115200);
  //  Serial.println("\nStarting\n\n");

  // Initialize TFT LCD display
  dm.begin();

  // Initialize keyboard splits data to defaults
  splits[0].channel  = splits[1].channel  = splits[2].channel  = DEFAULT_MIDI_CHANNEL;
  splits[0].velocity = splits[1].velocity = splits[2].velocity = DEFAULT_MIDI_VELOCITY;
  splits[0].patch    = splits[1].patch    = splits[2].patch    = DEFAULT_MIDI_PATCH;
  splits[0].keyboardShift = splits[1].keyboardShift = splits[2].keyboardShift
                            = DEFAULT_KEYBOARD_SHIFT;

  // Initialize pots to default CC values
  ccPot1 = DEFAULT_CCPOT_1;
  ccPot2 = DEFAULT_CCPOT_2;
  ccPot3 = DEFAULT_CCPOT_3;
  ccPot4 = DEFAULT_CCPOT_4;

  // Select the initially active split
  activeSplit = 0;

  // Startup hardware MIDI output port
  mi.begin();

  // Select acoustic grand piano from general MIDI
  mi.sendProgramChange(DEFAULT_MIDI_PATCH, DEFAULT_MIDI_CHANNEL);
}

/*********************************************************/
/***                    Program Loop                   ***/
/*********************************************************/
void loop(void) {

  // Update Control states
  controls.controlsUpdate();

  // Poll the keyboard for key events
  kb.pollKeyboard(processKeyboardEvent);

  // Run the Finite State Machine (FSM)
  fsm.run();
}
