/*
   MIDIIntfc.h - Wrapper for MIDI interface

   Concept, Design and Implementation by: Craig A. Lindley
   Last Update: 04/20/2022
*/

#ifndef MIDIINTFC_H
#define MIDIINTFC_H

#include "MIDI.h"

// Instantiate MIDI object for hardware output on GPIO pin 1
// MIDI data will also be output on USB
MIDI_CREATE_INSTANCE(HardwareSerial, Serial1, MIDI);

class MIDIIntfc {

  public:

    // Startup hardware MIDI output port
    void begin(void) {
      MIDI.begin();
    }

    // Send the appropriate MIDI note out
    // keyboard shift is -2, -1, 0, +1, +2 and controls the octave note is in
    void doMidiOutput(int channel, int velocity, int keyboardShift, int keyState, int keyNumber) {

      // Convert key number to midi note number
      int midiNoteNumber = keyNumber + DEFAULT_MIDI_NOTE_OFFSET + (keyboardShift * 12) - 1;

      if (keyState) {
        MIDI.sendNoteOn(midiNoteNumber, velocity, channel);
#ifdef ENABLE_USB_MIDI
        usbMIDI.sendNoteOn(midiNoteNumber, velocity, channel);
#endif
        Serial.printf("On Note #: %d, Vel: %d, Ch: %d\n",
                      midiNoteNumber, velocity, channel);
      } else  {
        MIDI.sendNoteOff(midiNoteNumber, velocity, channel);
#ifdef ENABLE_USB_MIDI
        usbMIDI.sendNoteOff(midiNoteNumber, velocity, channel);
#endif
        Serial.printf("Off Note #: %d, Vel: %d, Ch: %d\n",
                      midiNoteNumber, velocity, channel);
      }
    }

    // Send program/patch change MIDI message
    void sendProgramChange(int patch, int channel) {
      Serial.printf("ProgramChange: Patch: %d, Channel: %d\n", patch, channel);
      MIDI.sendProgramChange(patch, channel);
#ifdef ENABLE_USB_MIDI
      usbMIDI.sendProgramChange(patch, channel);
#endif
    }

    // Send control change MIDI message
    void sendControlChange(byte ccNumber, byte ccValue, int channel) {
      Serial.printf("ControlChange: CC: %d, CC Value: %d, Channel: %d\n", ccNumber, ccValue, channel);
      MIDI.sendControlChange(ccNumber, ccValue, channel);
#ifdef ENABLE_USB_MIDI
      usbMIDI.sendControlChange(ccNumber, ccValue, channel);
#endif
    }

    // Send pitch bend MIDI message
    void sendPitchBend(int value, int channel) {
      Serial.printf("Pitch Bend Value: %d, Channel: %d\n", value, channel);
      MIDI.sendPitchBend(value, channel);
#ifdef ENABLE_USB_MIDI
      usbMIDI.sendPitchBend(value, channel);
#endif
    }

  protected:

};

#endif
