/*
   Class for setting a Splits MIDI channel patch

   Runs its own FSM

   Concept, Design and Implementation by: Craig A. Lindley
   Last Update: 04/20/2022
*/

#ifndef PATCH_SELECTOR_H
#define PATCH_SELECTOR_H

class PatchSelector {

  public:

    PatchSelector() {}

    void init(int _splitNumber, Controls _controls, DisplayManager _dm, MIDIIntfc _mi) {

      // Save incoming
      splitNumber = _splitNumber;
      controls = _controls;
      dm = _dm;
      mi = _mi;

      patchCategoryEncoderValue = -1;;
      patchEncoderValue = -1;

      dm.eraseInfoAreaFromLine(4);
      dm.drawCenteredText("PatchCategory", 4, 2, ST77XX_GREEN);

      // Set initial state
      state = PATCH_CAT;
    }

    boolean run() {

      switch (state) {

        case PATCH_CAT:
          {
            // Rotate through the MIDI patch categories
            if (controls.readEncoder(&patchCategoryEncoderValue, 1, 0, 15)) {
              dm.eraseInfoAreaFromLine(5);

              // Format and display patch category
              sprintf(textBuffer, "%s", gm1PatchCategories[patchCategoryEncoderValue]);
              dm.drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
            }

            // Has a item been selected ?
            if (controls.encoderButtonSingleClicked()) {

              // Set next state
              state = PATCH_INIT;
            }
          }
          break;

        case PATCH_INIT:
          {
            dm.eraseInfoAreaFromLine(4);
            dm.drawCenteredText(textBuffer, 4, 2, ST77XX_GREEN);

            // Set next state
            state = PATCH;
          }
          break;

        case PATCH:
          {
            // Rotate through the MIDI patches
            if (controls.readEncoder(&patchEncoderValue, 1, 0, 7)) {
              dm.eraseInfoAreaFromLine(5);

              patchIndex = (patchCategoryEncoderValue * 8) + patchEncoderValue;

              // Format and display patch name
              sprintf(textBuffer, "%s", gm1Patches[patchIndex]);
              dm.drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
            }

            // Has a item been selected ?
            if (controls.encoderButtonSingleClicked()) {
              patchIndex++;

              splits[splitNumber].patch = patchIndex;

              mi.sendProgramChange(patchIndex, splits[splitNumber].channel);

              dm.eraseInfoAreaFromLine(4);

              return true;
            }
          }
      }
      return false;
    }

  protected:

    enum STATES { PATCH_CAT, PATCH_INIT, PATCH };
    enum STATES state;

    int splitNumber;
    Controls controls;
    DisplayManager dm;
    MIDIIntfc mi;

    int32_t patchCategoryEncoderValue;
    int32_t patchEncoderValue;
    int patchIndex;

    char textBuffer[30];

    // Array of patch categories
    const char *gm1PatchCategories[16] {
      "Pianos", "Ch Percussion", "Organs",
      "Guitars", "Basses", "Strings",
      "Ensembles", "Brass", "Reeds",
      "Pipes", "Synth Leads", "Synth Pads",
      "Synth Effects", "Ethnics", "Percussives",
      "Sound Effects"
    };

    // Array of General MIDI patches
    const char *gm1Patches[128] {
      // Patches 1-8: Piano
      "Acoust Grand", "Bright Acoust", "Elec Grand",
      "Honky Tonk", "Electric 1", "Electric 2",
      "Harpsichord", "Clavi",

      // Patches 9-16: Chromatic Percussion
      "Celesta", "Glockenspiel", "Music Box",
      "Vibraphone", "Marimba", "Xylophone",
      "Tubular Bells", "Dulcimer",

      // Patches 17-24: Organ
      "Drawbar", "Percussive", "Rock",
      "Church", "Reed", "Accordion",
      "Harmonica", "Tango Accord",

      // Patches 25-32: Guitar
      "Acoust Nylon", "Acoust Steel", "Elect Jazz",
      "Elect Clean", "Elect Muted", "Overdriven",
      "Distortion", "Harmonics",

      // Patches 33-40: Bass
      "Acoustic Bass", "Elect Finger", "Elect Pick",
      "Fretless", "Slap 1", "Slap 2",
      "Synth 1", "Synth 2",

      // Patches 41-48: Strings
      "Violin", "Viola", "Cello",
      "Contrabass", "Trem Strings", "Pizzi Strings",
      "Orch Harp", "Timpani",

      // Patches 49-56: Ensemble
      "Strings 1", "Strings 2", "SynthStrings1",
      "SynthStrings2", "Choir Aahs", "Voice Oohs",
      "Synth Voice", "Orchestra Hit",

      // Patches 57-64: Brass
      "Trumpet", "Trombone", "Tuba",
      "Muted Trumpet", "French Horn", "Brass Section",
      "SynthBrass 1", "SynthBrass 2",

      // Patches 65-72: Reed
      "Soprano Sax", "Alto Sax", "Tenor Sax",
      "Baritone Sax", "Oboe", "English Horn",
      "Bassoon", "Clarinet",

      // Patches 73-80: Pipe
      "Piccolo", "Flute", "Recorder",
      "Pan Flute", "Blown Bottle", "Shakuhachi",
      "Whistle", "Ocarina",

      // Patches 81-88: Synth Lead
      "Square", "Sawtooth", "Calliope",
      "Chiff", "Lead Charang", "Lead Voice",
      "Lead Fifths", "Bass + Lead",

      // Patches 89-96: Synth Pad
      "New age", "Warm", "Polysynth",
      "Choir", "Bowed", "Metallic",
      "Halo", "Sweep",

      // Patches 97-104: Synth Effects
      "Rain", "Soundtrack", "Crystal",
      "Atmosphere", "Brightness", "Goblins",
      "Echoes", "SciFi",

      // Patches 105-112: Ethnic
      "Sitar", "Banjo", "Shamisen",
      "Koto", "Kalimba", "Bag pipe",
      "Fiddle", "Shanai",

      // Patches 113-120: Percussive
      "Tinkle Bell", "Agogo", "Steel Drums",
      "Woodblock", "Taiko Drum", "Melodic Tom",
      "Synth Drum", "Reverse Cymbal",

      // Patches 121-128: Sound Effects
      "Fret Noise", "Breath Noise", "Seashore",
      "Bird Tweet", "Telephone", "Helicopter",
      "Applause", "Gunshot"
    };
};

#endif
