/*
   HMSynth Keyboard State Machine

   Concept, Design and Implementation by: Craig A. Lindley
   Last Update: 04/26/2022
*/

#ifndef STATEMACHINE_H
#define STATEMACHINE_H

// Software restart data used to reboot Teensy 3.6
#define CPU_RESTART_ADDR (uint32_t *)0xE000ED0C
#define CPU_RESTART_VAL 0x5FA0004
#define CPU_RESTART (*CPU_RESTART_ADDR = CPU_RESTART_VAL);

// CC values associated with the 4 pots
int ccPot1;
int ccPot2;
int ccPot3;
int ccPot4;

// Data structure for CC messages
typedef struct {
  int ccCode;
  const char *ccString;
} CC_DATA;

int activeSplit;
boolean portamentoOn;

char textBuffer[30];

class StateMachine {
  public:

    // Class constructor
    StateMachine(Controls _controls, DisplayManager _dm,
                 Keyboard _kb, MIDIIntfc _mi,
                 ChannelSelector _cs, VelocitySelector _vs,
                 KeyShiftSelector _kss, PatchSelector _ps) {

      // Save incoming
      controls = _controls;
      dm = _dm;
      kb = _kb;
      mi = _mi;
      cs = _cs;
      vs = _vs;
      kss = _kss;
      ps = _ps;

      // Set initial state
      state = STARTUP;
    }

    // Turn all on notes off
    void allNotesOff(int channel, int keyboardShift) {

      // Look at each key 1 .. 36 in history
      for (int key = 1; key < 37; key++) {
        // Is this key on ?
        if (kb.getKeyState(key)) {
          // Yes it was on turn it off
          mi.doMidiOutput(channel, 0, keyboardShift, false, key);
        }
      }
      kb.clearKeyHistory();
    }

    void run() {

      switch (state) {
        case STARTUP:
          {
            // Do pre-read of pots to establish initial values
            controls.modWheelActive();
            controls.pitchWheelActive();
            controls.pot1Active();
            controls.pot2Active();
            controls.pot3Active();
            controls.pot4Active();

            // Set encoder value defaults
            menuEncoderValue = -1;
            splitEncoderValue = -1;
            pot1EncoderValue = 0;
            pot2EncoderValue = 0;
            pot3EncoderValue = 0;
            pot4EncoderValue = 0;

            // Portamento is off by default
            // Portamento can be set independently for each split
            portamentoOn = false;

            // Next state
            state = CHECK_CONTROLS_INIT;
          }
          break;

        case CHECK_CONTROLS_INIT:
          {
            // Format LCD for info display
            dm.formatLCD();

            // Set next state
            state = CHECK_CONTROLS;
          }
          break;

        case CHECK_CONTROLS:
          {
            // Multi function button single click is portamento On/Off
            if (controls.multiFunctionButtonSingleClicked()) {
              dm.eraseInfoAreaFromLine(4);
              if (portamentoOn == false) {
                portamentoOn = true;
                mi.sendControlChange(PORTAMENTO_CC, 127, splits[activeSplit].channel);
                dm.drawCenteredText("Portamento", 4, 2, ST77XX_GREEN);
                dm.drawCenteredText("On", 5, 2, ST77XX_GREEN);

              } else {
                portamentoOn = false;
                mi.sendControlChange(PORTAMENTO_CC, 0, splits[activeSplit].channel);
                dm.drawCenteredText("Portamento", 4, 2, ST77XX_GREEN);
                dm.drawCenteredText("Off", 5, 2, ST77XX_GREEN);
              }
            }

            // Multi function button double click is all notes off
            if (controls.multiFunctionButtonDoubleClicked()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("All notes off", 4, 2, ST77XX_GREEN);

              // Splits are enabled so send CC on each split channel
              mi.sendControlChange(120, 0, splits[0].channel);
              mi.sendControlChange(120, 0, splits[1].channel);
              mi.sendControlChange(120, 0, splits[2].channel);
              kb.clearKeyHistory();
            }

            // Multi function button long click is reset
            if (controls.multiFunctionButtonLongClicked()) {
              // Reboot the Teensy 3.6
              CPU_RESTART
            }

            // Check Up button single click - raise octave
            if (controls.upButtonSingleClicked()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Octave Up", 4, 2, ST77XX_GREEN);

              allNotesOff(splits[activeSplit].channel, splits[activeSplit].keyboardShift);
              splits[activeSplit].keyboardShift++;
              if (splits[activeSplit].keyboardShift > 2) {
                splits[activeSplit].keyboardShift = 2;
              }
            }

            // Check Down button single click - lower octave
            if (controls.downButtonSingleClicked()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Octave Down", 4, 2, ST77XX_GREEN);

              allNotesOff(splits[activeSplit].channel, splits[activeSplit].keyboardShift);
              splits[activeSplit].keyboardShift--;
              if (splits[activeSplit].keyboardShift < -2) {
                splits[activeSplit].keyboardShift = -2;
              }
            }

            // Check Up button double click - increment patch
            if (controls.upButtonDoubleClicked()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Patch +", 4, 2, ST77XX_GREEN);

              allNotesOff(splits[activeSplit].channel, splits[activeSplit].keyboardShift);
              splits[activeSplit].patch++;
              if (splits[activeSplit].patch > 127) {
                splits[activeSplit].patch = 127;
              }
              mi.sendProgramChange(splits[activeSplit].patch, splits[activeSplit].channel);
            }

            // Check Down button double click - decrement patch
            if (controls.downButtonDoubleClicked()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Patch -", 4, 2, ST77XX_GREEN);
              allNotesOff(splits[activeSplit].channel, splits[activeSplit].keyboardShift);

              splits[activeSplit].patch--;

              if (splits[activeSplit].patch < 0) {
                splits[activeSplit].patch = 0;
              }
              mi.sendProgramChange(splits[activeSplit].patch, splits[activeSplit].channel);
            }

            // Is pot active ?
            if (controls.pot1Active()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Pot1", 4, 2, ST77XX_GREEN);

              // Pot is active, get new value
              value = controls.pot1Value();
              mi.sendControlChange(ccPot1, value, splits[activeSplit].channel);
            }

            // Is pot active ?
            if (controls.pot2Active()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Pot2", 4, 2, ST77XX_GREEN);

              // Pot is active, get new value
              value = controls.pot2Value();
              mi.sendControlChange(ccPot2, value, splits[activeSplit].channel);
            }

            // Is pot active ?
            if (controls.pot3Active()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Pot3", 4, 2, ST77XX_GREEN);

              // Pot is active, get new value
              value = controls.pot3Value();
              mi.sendControlChange(ccPot3, value, splits[activeSplit].channel);
            }

            // Is pot active ?
            if (controls.pot4Active()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Pot4", 4, 2, ST77XX_GREEN);

              // Pot is active, get new value
              value = controls.pot4Value();
              mi.sendControlChange(ccPot4, value, splits[activeSplit].channel);
            }

            // Is mod wheel active ?
            if (controls.modWheelActive()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Mod Wheel", 4, 2, ST77XX_GREEN);

              // Pot is active, get new value
              value = controls.modWheelValue();

              int newValue = 0;

              // Map mod wheel value into proper range
              if ((value >= 490) && (value <= 540)) {
                // Joystick/mod wheel is centered
                newValue = 0;
              } else if (value < 490) {
                newValue = map(value, 490, 0, 0, 127);
              } else {
                newValue = map(value, 540, 1023, 0, 127);
              }
              Serial.println(newValue);
              mi.sendControlChange(1, newValue, splits[activeSplit].channel);
            }

            // Is pitch wheel active ?
            if (controls.pitchWheelActive()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Pitch Wheel", 4, 2, ST77XX_GREEN);

              // Pot is active, get new value
              value = controls.pitchWheelValue();

              int newValue = 0;

              // Map mod wheel value into proper range
              if ((value >= 490) && (value <= 540)) {
                // Joystick/mod wheel is centered
                newValue = 0;
              } else if (value < 490) {
                newValue = map(value, 490, 0, 0, -8192);
              } else {
                newValue = map(value, 540, 1023, 0, 8191);
              }
              Serial.println(newValue);
              mi.sendPitchBend(newValue, splits[activeSplit].channel);
            }

            // Rotary encoder single clicked ?
            if (controls.encoderButtonSingleClicked()) {
              // Yes set next state
              state = MENU_INIT;
            }

            // Rotary encoder long clicked ?
            if (controls.encoderButtonLongClicked()) {
              dm.eraseInfoAreaFromLine(4);
              dm.drawCenteredText("Splits", 4, 2, ST77XX_GREEN);
              dm.drawCenteredText("Channels", 5, 2, ST77XX_GREEN);
              dm.drawCenteredText("Set", 6, 2, ST77XX_GREEN);

              // Set the channel for each of the three splits to default
              splits[0].channel = 1;
              splits[1].channel = 2;
              splits[2].channel = 3;
            }
          }
          break;

        case MENU_INIT:
          {
            dm.eraseInfoAreaFromLine(4);
            dm.drawCenteredText("Configuration", 4, 2, ST77XX_GREEN);

            // Set next state
            state = MENU;
          }
          break;

        case MENU:
          {
            // Rotate through all high level menu items
            if (controls.readEncoder(&menuEncoderValue, 1, 0, NUMBER_HL_MENU_ITEMS - 1)) {
              dm.eraseInfoAreaFromLine(5);
              dm.drawCenteredText(hlMenu[menuEncoderValue], 6, 2, ST77XX_GREEN);
            }

            // Has a menu item been selected ?
            if (controls.encoderButtonSingleClicked()) {
              switch (menuEncoderValue) {
                case 0:
                  state = SPLIT1_INIT;
                  break;
                case 1:
                  state = SPLIT2_INIT;
                  break;
                case 2:
                  state = SPLIT3_INIT;
                  break;
                case 3:
                  state = POT1_INIT;
                  break;
                case 4:
                  state = POT2_INIT;
                  break;
                case 5:
                  state = POT3_INIT;
                  break;
                case 6:
                  state = POT4_INIT;
                  break;
                case 7:
                  state = RESET;
                  break;
              }
            }

            // Has encoder been double clicked ?
            if (controls.encoderButtonDoubleClicked()) {
              // Set next state
              state = CHECK_CONTROLS_INIT;
            }
          }
          break;

        case SPLIT1_INIT:
          {
            dm.eraseInfoAreaFromLine(4);
            dm.drawCenteredText("Split1 Config", 4, 2, ST77XX_GREEN);

            // Set next state
            state = SPLIT1;
          }
          break;

        case SPLIT1:
          {
            // Rotate through the split menu
            if (controls.readEncoder(&splitEncoderValue, 1, 0, NUMBER_SPLIT_MENU_ITEMS - 1)) {
              dm.eraseInfoAreaFromLine(5);

              // Format and display split menu option
              sprintf(textBuffer, "%s", splitMenu[splitEncoderValue]);
              dm.drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
            }

            // Has a item been selected ?
            if (controls.encoderButtonSingleClicked()) {

              switch (splitEncoderValue) {
                case 0:
                  state = SPLIT1_CHANNEL_INIT;
                  break;
                case 1:
                  state = SPLIT1_VELOCITY_INIT;
                  break;
                case 2:
                  state = SPLIT1_KEYSHIFT_INIT;
                  break;
                case 3:
                  state = SPLIT1_PATCH_INIT;
                  break;
              }
            }

            // Has encoder been double clicked ?
            if (controls.encoderButtonDoubleClicked()) {
              // Set next state
              state = CHECK_CONTROLS_INIT;
            }

          }
          break;

        case SPLIT1_CHANNEL_INIT:
          {
            cs.init(0, controls, dm);

            // Set next state
            state = SPLIT1_CHANNEL;
          }
          break;

        case SPLIT1_CHANNEL:
          {
            boolean done = cs.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT1_INIT;
            }
          }
          break;

        case SPLIT1_VELOCITY_INIT:
          {
            vs.init(0, controls, dm);

            // Set next state
            state = SPLIT1_VELOCITY;
          }
          break;

        case SPLIT1_VELOCITY:
          {
            boolean done = vs.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT1_INIT;
            }
          }
          break;

        case SPLIT1_KEYSHIFT_INIT:
          {
            kss.init(0, controls, dm);

            // Set next state
            state = SPLIT1_KEYSHIFT;

          }
          break;

        case SPLIT1_KEYSHIFT:
          {
            boolean done = kss.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT1_INIT;
            }
          }
          break;

        case SPLIT1_PATCH_INIT:
          {
            ps.init(0, controls, dm, mi);

            // Set next state
            state = SPLIT1_PATCH;
          }
          break;

        case SPLIT1_PATCH:
          {
            boolean done = ps.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT1_INIT;
            }
          }
          break;

        case SPLIT2_INIT:
          {
            dm.eraseInfoAreaFromLine(4);
            dm.drawCenteredText("Split2 Config", 4, 2, ST77XX_GREEN);

            // Set next state
            state = SPLIT2;
          }
          break;

        case SPLIT2:
          {
            // Rotate through the split menu
            if (controls.readEncoder(&splitEncoderValue, 1, 0, NUMBER_SPLIT_MENU_ITEMS - 1)) {
              dm.eraseInfoAreaFromLine(5);

              // Format and display split menu option
              sprintf(textBuffer, "%s", splitMenu[splitEncoderValue]);
              dm.drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
            }

            // Has a item been selected ?
            if (controls.encoderButtonSingleClicked()) {

              switch (splitEncoderValue) {
                case 0:
                  state = SPLIT2_CHANNEL_INIT;
                  break;
                case 1:
                  state = SPLIT2_VELOCITY_INIT;
                  break;
                case 2:
                  state = SPLIT2_KEYSHIFT_INIT;
                  break;
                case 3:
                  state = SPLIT2_PATCH_INIT;
                  break;
              }
            }

            // Has encoder been double clicked ?
            if (controls.encoderButtonDoubleClicked()) {
              // Set next state
              state = CHECK_CONTROLS_INIT;
            }
          }
          break;

        case SPLIT2_CHANNEL_INIT:
          {
            cs.init(1, controls, dm);

            // Set next state
            state = SPLIT2_CHANNEL;
          }
          break;

        case SPLIT2_CHANNEL:
          {
            boolean done = cs.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT2_INIT;
            }
          }
          break;

        case SPLIT2_VELOCITY_INIT:
          {
            vs.init(1, controls, dm);

            // Set next state
            state = SPLIT2_VELOCITY;
          }
          break;

        case SPLIT2_VELOCITY:
          {
            boolean done = vs.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT2_INIT;
            }
          }
          break;

        case SPLIT2_KEYSHIFT_INIT:
          {
            kss.init(1, controls, dm);

            // Set next state
            state = SPLIT2_KEYSHIFT;

          }
          break;

        case SPLIT2_KEYSHIFT:
          {
            boolean done = kss.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT2_INIT;
            }
          }
          break;

        case SPLIT2_PATCH_INIT:
          {
            ps.init(1, controls, dm, mi);

            // Set next state
            state = SPLIT2_PATCH;
          }
          break;

        case SPLIT2_PATCH:
          {
            boolean done = ps.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT2_INIT;
            }
          }
          break;

        case SPLIT3_INIT:
          {
            dm.eraseInfoAreaFromLine(4);
            dm.drawCenteredText("Split3 Config", 4, 2, ST77XX_GREEN);

            // Set next state
            state = SPLIT3;
          }
          break;

        case SPLIT3:
          {
            // Rotate through the split menu
            if (controls.readEncoder(&splitEncoderValue, 1, 0, NUMBER_SPLIT_MENU_ITEMS - 1)) {
              dm.eraseInfoAreaFromLine(5);

              // Format and display split menu option
              sprintf(textBuffer, "%s", splitMenu[splitEncoderValue]);
              dm.drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
            }

            // Has a item been selected ?
            if (controls.encoderButtonSingleClicked()) {

              switch (splitEncoderValue) {
                case 0:
                  state = SPLIT3_CHANNEL_INIT;
                  break;
                case 1:
                  state = SPLIT3_VELOCITY_INIT;
                  break;
                case 2:
                  state = SPLIT3_KEYSHIFT_INIT;
                  break;
                case 3:
                  state = SPLIT3_PATCH_INIT;
                  break;
              }
            }

            // Has encoder been double clicked ?
            if (controls.encoderButtonDoubleClicked()) {
              // Set next state
              state = CHECK_CONTROLS_INIT;
            }

          }
          break;

        case SPLIT3_CHANNEL_INIT:
          {
            cs.init(2, controls, dm);

            // Set next state
            state = SPLIT3_CHANNEL;
          }
          break;

        case SPLIT3_CHANNEL:
          {
            boolean done = cs.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT3_INIT;
            }
          }
          break;

        case SPLIT3_VELOCITY_INIT:
          {
            vs.init(2, controls, dm);

            // Set next state
            state = SPLIT3_VELOCITY;
          }
          break;

        case SPLIT3_VELOCITY:
          {
            boolean done = vs.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT3_INIT;
            }
          }
          break;

        case SPLIT3_KEYSHIFT_INIT:
          {
            kss.init(2, controls, dm);

            // Set next state
            state = SPLIT3_KEYSHIFT;
          }
          break;

        case SPLIT3_KEYSHIFT:
          {
            boolean done = kss.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT3_INIT;
            }
          }
          break;

        case SPLIT3_PATCH_INIT:
          {
            ps.init(2, controls, dm, mi);

            // Set next state
            state = SPLIT3_PATCH;
          }
          break;

        case SPLIT3_PATCH:
          {
            boolean done = ps.run();
            if (done) {
              dm.formatLCD();

              // Set next state
              state = SPLIT3_INIT;
            }
          }
          break;

        case POT1_INIT:
          {
            dm.eraseInfoAreaFromLine(4);
            dm.drawCenteredText("Pot1 CC", 4, 2, ST77XX_GREEN);

            // Set next state
            state = POT1;
          }
          break;

        case POT1:
          {
            // Rotate through the CC values 0 .. 55
            if (controls.readEncoder(&pot1EncoderValue, 1, 0, 55)) {
              dm.eraseInfoAreaFromLine(5);

              ccd = midiCC[pot1EncoderValue];

              // Format and display key offset number
              sprintf(textBuffer, "%s", ccd.ccString);
              dm.drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
            }

            // Has a item been selected ?
            if (controls.encoderButtonSingleClicked()) {

              ccPot1 = ccd.ccCode;

              // Set next state
              state = MENU_INIT;
            }

            // Has encoder been double clicked ?
            if (controls.encoderButtonDoubleClicked()) {
              // Set next state
              state = CHECK_CONTROLS_INIT;
            }
          }
          break;

        case POT2_INIT:
          {
            dm.eraseInfoAreaFromLine(4);
            dm.drawCenteredText("Pot2 CC", 4, 2, ST77XX_GREEN);

            // Set next state
            state = POT2;
          }
          break;

        case POT2:
          {
            // Rotate through the CC values 0 .. 55
            if (controls.readEncoder(&pot2EncoderValue, 1, 0, 55)) {
              dm.eraseInfoAreaFromLine(5);

              ccd = midiCC[pot2EncoderValue];

              // Format and display key offset number
              sprintf(textBuffer, "%s", ccd.ccString);
              dm.drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
            }

            // Has a item been selected ?
            if (controls.encoderButtonSingleClicked()) {

              ccPot2 = ccd.ccCode;

              // Set next state
              state = MENU_INIT;
            }

            // Has encoder been double clicked ?
            if (controls.encoderButtonDoubleClicked()) {
              // Set next state
              state = CHECK_CONTROLS_INIT;
            }
          }
          break;

        case POT3_INIT:
          {
            dm.eraseInfoAreaFromLine(4);
            dm.drawCenteredText("Pot3 CC", 4, 2, ST77XX_GREEN);

            // Set next state
            state = POT3;
          }
          break;

        case POT3:
          {
            // Rotate through the CC values 0 .. 55
            if (controls.readEncoder(&pot3EncoderValue, 1, 0, 55)) {
              dm.eraseInfoAreaFromLine(5);

              ccd = midiCC[pot3EncoderValue];

              // Format and display key offset number
              sprintf(textBuffer, "%s", ccd.ccString);
              dm.drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
            }

            // Has a item been selected ?
            if (controls.encoderButtonSingleClicked()) {

              ccPot3 = ccd.ccCode;

              // Set next state
              state = MENU_INIT;
            }

            // Has encoder been double clicked ?
            if (controls.encoderButtonDoubleClicked()) {
              // Set next state
              state = CHECK_CONTROLS_INIT;
            }
          }
          break;

        case POT4_INIT:
          {
            dm.eraseInfoAreaFromLine(4);
            dm.drawCenteredText("Pot4 CC", 4, 2, ST77XX_GREEN);

            // Set next state
            state = POT4;
          }
          break;

        case POT4:
          {
            // Rotate through the CC values 0 .. 55
            if (controls.readEncoder(&pot4EncoderValue, 1, 0, 55)) {
              dm.eraseInfoAreaFromLine(5);

              ccd = midiCC[pot4EncoderValue];

              // Format and display key offset number
              sprintf(textBuffer, "%s", ccd.ccString);
              dm.drawCenteredText(textBuffer, 6, 2, ST77XX_GREEN);
            }

            // Has a item been selected ?
            if (controls.encoderButtonSingleClicked()) {

              ccPot4 = ccd.ccCode;

              // Set next state
              state = MENU_INIT;
            }

            // Has encoder been double clicked ?
            if (controls.encoderButtonDoubleClicked()) {
              // Set next state
              state = CHECK_CONTROLS_INIT;
            }
          }
          break;

        case RESET:
          {
            // Reboot the Teensy 3.6
            CPU_RESTART
          }
          break;
      }
    }

  protected:

    uint16_t value;

    int32_t splitEncoderValue;
    int32_t menuEncoderValue;
    int32_t pot1EncoderValue;
    int32_t pot2EncoderValue;
    int32_t pot3EncoderValue;
    int32_t pot4EncoderValue;

    CC_DATA ccd;

    Controls controls;
    DisplayManager dm;
    Keyboard kb;
    MIDIIntfc mi;
    ChannelSelector cs;
    VelocitySelector vs;
    KeyShiftSelector kss;
    PatchSelector ps;

    // Finite State Machine states
    enum STATES {
      STARTUP, CHECK_CONTROLS_INIT, CHECK_CONTROLS,
      MENU_INIT, MENU,

      SPLIT1_INIT, SPLIT1,
      SPLIT1_CHANNEL_INIT, SPLIT1_CHANNEL,
      SPLIT1_VELOCITY_INIT, SPLIT1_VELOCITY,
      SPLIT1_KEYSHIFT_INIT, SPLIT1_KEYSHIFT,
      SPLIT1_PATCH_INIT, SPLIT1_PATCH,

      SPLIT2_INIT, SPLIT2,
      SPLIT2_CHANNEL_INIT, SPLIT2_CHANNEL,
      SPLIT2_VELOCITY_INIT, SPLIT2_VELOCITY,
      SPLIT2_KEYSHIFT_INIT, SPLIT2_KEYSHIFT,
      SPLIT2_PATCH_INIT, SPLIT2_PATCH,

      SPLIT3_INIT, SPLIT3,
      SPLIT3_CHANNEL_INIT, SPLIT3_CHANNEL,
      SPLIT3_VELOCITY_INIT, SPLIT3_VELOCITY,
      SPLIT3_KEYSHIFT_INIT, SPLIT3_KEYSHIFT,
      SPLIT3_PATCH_INIT, SPLIT3_PATCH,

      POT1_INIT, POT1,
      POT2_INIT, POT2,
      POT3_INIT, POT3,
      POT4_INIT, POT4,
      RESET
    };

    // Current state of FSM
    STATES state;

    // High/Top level menu items
    const char *hlMenu[8] {
      "Split 1", "Split 2", "Split 3",
      "Pot1 CC", "Pot2 CC",
      "Pot3 CC", "Pot4 CC",
      "Reset"
    };
    const int NUMBER_HL_MENU_ITEMS = sizeof(hlMenu) / sizeof(const char *);

    // Split submenu items
    const char *splitMenu[4] {
      "Channel", "Velocity", "Key Shift", "Patch"
    };
    const int NUMBER_SPLIT_MENU_ITEMS = sizeof(splitMenu) / sizeof(const char *);

    // Array of control change messages
    CC_DATA midiCC[56] {
      {0, "Bank Select"}, {1, "Modulation"}, {2, "Breath Ctlr"},
      {4, "Foot Ctlr"}, {5, "Portamento Tm"},
      {6, "DataEntry MSB"}, {7, "Volume"}, {8, "Balance"},
      {10, "Pan"}, {11, "Expression"}, {12, "Effect Ctl 1"},
      {13, "Effect Ctl 2"}, {16, "GP1"}, {17, "GP2"},
      {18, "GP3"}, {19, "GP4"}, {64, "D/S Pedal"},
      {65, "Portamento"}, {66, "Sostenuto"}, {67, "Soft Pedal"},
      {68, "Legato"}, {69, "Hold 2"}, {70, "Sound Ctrl 1"},
      {71, "Sound Ctrl 2"}, {72, "Sound Ctrl 3"}, {73, "Sound Ctrl 4"},
      {74, "Sound Ctrl 5"}, {75, "Sound Ctrl 6"}, {76, "Sound Ctrl 7"},
      {77, "Sound Ctrl 8"}, {78, "Sound Ctrl 9"}, {79, "Sound Ctrl 10"},
      {80, "Gen1 On/Off"}, {81, "Gen2 On/Off"}, {82, "Gen3 On/Off"},
      {83, "Gen4 On/Off"}, {84, "Portamento"}, {91, "Effect1 Depth"},
      {92, "Effect2 Depth"}, {93, "Effect3 Depth"}, {94, "Effect4 Depth"},
      {95, "Effect5 Depth"}, {96, "Data Inc"}, {97, "Data Dec"},
      {98, "NRPN LSB"}, {99, "NRPN MSB"}, {100, "RPN LSB"},
      {101, "RPN MSB"}, {120, "All Sound Off"}, {121, "Reset Ctrls"},
      {122, "Local On/Off"}, {123, "All Notes Off"}, {124, "OMNI Off"},
      {125, "OMNI On"}, {126, "Mono Mode"}, {127, "Poly Mode"}
    };
};

#endif
